/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 * The following code initiates various threads which are used to read messages from named pipes
 * as also to sent those messages to the broker
 */

#include <unistd.h>
#include <iostream>    
#include <fstream>   
#include <stdlib.h>   
#include <vector>
#include <decaf/lang/Thread.h>
#include <decaf/lang/Runnable.h>
#include <decaf/util/concurrent/CountDownLatch.h>
#include <decaf/lang/Long.h>
#include <decaf/util/Date.h>
#include <activemq/core/ActiveMQConnectionFactory.h>
#include <activemq/util/Config.h>
#include <activemq/library/ActiveMQCPP.h>
#include <cms/Connection.h>
#include <cms/Session.h>
#include <cms/TextMessage.h>
#include <cms/BytesMessage.h>
#include <cms/MapMessage.h>
#include <cms/ExceptionListener.h>
#include <cms/MessageListener.h>
#include <string>
#include <sstream>
#include <iomanip>
#include <string.h>
#include <errno.h>
#include <iostream>
#include "MsgPipe.h"
#include "MsgProducer.h"
#include "concurrent_queue.h"
#include "half_duplex.h"
#include "utility_routines.h"

using namespace std;


int main() {

    daemon(0, 0);  
   
    //initialize here to avoid race conditions  
    concurrent_queue::getInstance();
      
    MsgPipe pipeMsg1(HALF_DUPLEX);
    MsgPipe pipeMsg2(HALF_DUPLEX2); 
    MsgPipe pipeMsg3(HALF_DUPLEX3); 
    MsgProducer producer;
  
    
    // Start the producer thread.
    Thread pipeThread(&pipeMsg1);
    pipeThread.start();

       // Start the producer thread.
    Thread pipeThread2(&pipeMsg2);
    pipeThread2.start();

    Thread pipeThread3(&pipeMsg3);
    pipeThread3.start();


    // Start the producer thread.
    Thread producerThread(&producer);
    producerThread.start();

 
    // Wait for the threads to complete.
    pipeThread.join();    
    pipeThread2.join();
    pipeThread3.join();
    producerThread.join();
    
    pipeMsg1.cleanup();
    pipeMsg2.cleanup();    
    pipeMsg3.cleanup();    
    producer.cleanup();
    
    return 0;
}


