/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 * various utility routines to be used across libraries and executables
 */

#include "utility_routines.h"
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include<cctype> //for isdigit
#include<sstream> //to phrase string
#include<string>
#include "Logger.h"
#include <sys/stat.h>
#include <algorithm>
#include "half_duplex.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>

#include "compress.h"


#define MILLI 36000000

const std::string CFG_FILE_NAME = "fts-msg-monitoring.conf";
const std::string CFG_FILE_PATH = "/etc/";
const std::string DATABASE_FILE_NAME = ".properties.xml";
const std::string DATABASE_FILE_PATH = "/etc/glite-data-transfer-agents.d/";

std::string BROKER;
std::string START;
std::string COMPLETE;
std::string CRON;
std::string TTL;
bool TOPIC;
bool ACTIVE;
bool ENABLELOG;
std::string LOGFILEDIR;
std::string LOGFILENAME;
std::string CRONFQDN;
bool ENABLEMSGLOG;

std::string USERNAME;
std::string PASSWORD;
bool USE_BROKER_CREDENTIALS;

//oracle credentials
static std::vector<std::string> fileDB;

//msg config file
static map<string, string> cfg;

//recovery vector in case no data is retrieved
static vector<std::string> recoveryVector(3, "");

//store fts endpoint for usage in the message field
static map<string, string> ftsendpoint;

std::string _getTimestamp() {
    time_t msec_started = time(NULL) * 1000; //the number of milliseconds since the epoch.
    std::string transfer_started_to_string = _to_string<long double>(msec_started, std::dec);
    return transfer_started_to_string;
}

std::string get_channel_(std::string tr_id) {
    size_t found = tr_id.find("__");
    if (found != string::npos)
        return tr_id.substr(0, (found));
}

std::string get_hostname(std::string hostname) {
    size_t found = 0;
    size_t found_colon = 0;
    size_t found_extra_slash = 0; //for gsiftp only

    if (hostname.compare(0, 9, "gsiftp://") == 0) {
        std::string temp = hostname.substr(9, (int(found) - 9));
        found = temp.find_first_of('/');
        if (found != string::npos) {
            found_extra_slash = temp.find_first_of('/');
            if (found_extra_slash != string::npos)
                return temp.substr(0, int(found_extra_slash));
            found_colon = temp.find_first_of(':');
            if (found_colon != string::npos)
                return temp.substr(0, int(found_colon));
            else
                return temp.substr(0, int(found));
        } else {
            found = temp.find_first_of(':');
            if (found != string::npos)
                return temp.substr(0, int(found));
        }
    }


    if (hostname.compare(0, 6, "srm://") == 0) {
        std::string temp = hostname.substr(6, (int(found) - 6));
        found = temp.find_first_of('/');
        if (found != string::npos) {
            found_colon = temp.find_first_of(':');
            if (found_colon != string::npos)
                return temp.substr(0, int(found_colon));
            else
                return temp.substr(0, int(found));
        } else {
            found = temp.find_first_of(':');
            if (found != string::npos)
                return temp.substr(0, int(found));
        }
    }
    return "invalid hostname";
}

std::string getUserName(std::string & value, std::vector<std::string>::iterator it) {
    size_t found = value.find("User");
    std::string username;
    if (found != std::string::npos) {
        username = *(++it);
        size_t val = username.find("</value>");
        return username.substr(7, (username.length() - 15));
    }
    return "";
}

std::string getPassword(std::string & value, std::vector<std::string>::iterator it) {
    size_t found = value.find("Password");
    std::string password;
    if (found != std::string::npos) {
        password = *(++it);
        size_t val = password.find("</value>");
        return password.substr(7, (password.length() - 15));
    }
    return "";
}

std::string getConnectString(std::string & value, std::vector<std::string>::iterator it) {
    size_t found = value.find("ConnectString");
    std::string connectstring("");
    std::string connectStringConcat("");
    size_t val;
    string::iterator itr;
    if (found != std::string::npos) {
        connectstring = *(++it);
        val = connectstring.find("</value>");
	if (val != std::string::npos) 
        	return connectstring.substr(7, (connectstring.length() - 15));
	else
	{
	  --it;
	  while( 1 )
	  {
	    connectStringConcat += *(++it);
	    val = connectStringConcat.find("</value>");
	    if (val != std::string::npos)
	    {	     
  	      connectStringConcat.erase(0,7);
	      connectStringConcat.erase(connectStringConcat.length()-8, connectStringConcat.length());	           
	      break;
	    } 
	  }
	}
    }
    return connectStringConcat;
}

int fexists(const char *filename) {
    struct stat buffer;
    if (stat(filename, &buffer)) return 1;
    return 0;
}

int getdir(string rootDir, vector<string> &files) {
    DIR *dir; //the directory
    struct dirent *dp;

    //open the directory
    if ((dir = opendir(rootDir.c_str())) == NULL) {
        //std::string errorMessage = "Error opening file, check GLITE_LOCATION: " + rootDir;
        //logger::writeLog(errorMessage, true);
        return errno;
    }

    while ((dp = readdir(dir)) != NULL) {
        files.push_back(string(dp->d_name));
    }

    closedir(dir);
    return 0;
}

std::string filesStore(const char* filename, const char *path, char *env) {
    std::string dirname;
    std::string rootDir;
    vector<std::string> files = vector<std::string > ();
    string inventory[3] = {"", "/usr", "/opt/glite"};
    std::string xmlfile;
    size_t found;

    try {
        if (env == NULL) {
            for (int i = 0; i < 3; ++i) {
                dirname = inventory[i];
                dirname = dirname + path;
                rootDir = std::string(dirname);
                getdir(rootDir, files);

                for (unsigned int i = 0; i < files.size(); i++) {
                     xmlfile = files[i];
                     found = xmlfile.find(filename);		     	    
                    if (found != std::string::npos) {		       
                        return dirname + xmlfile;
                    }
                }
            }
        } else {
            dirname = env;
            dirname = dirname + path;
            rootDir = std::string(dirname);
            getdir(rootDir, files);

            for (unsigned int i = 0; i < files.size(); i++) {
                xmlfile = files[i];
                found = xmlfile.find(filename);
                if (found != std::string::npos) {
                    return dirname + xmlfile;
                }
            }
            for (int i = 0; i < 3; ++i) {
                dirname = inventory[i];
                dirname = dirname + path;
                rootDir = std::string(dirname);
                getdir(rootDir, files);

                for (unsigned int i = 0; i < files.size(); i++) {
                    xmlfile = files[i];
                    found = xmlfile.find(filename);
                    if (found!=string::npos) {
                        return dirname + xmlfile;
                    }
                }
            }	    
        }
        return "";
    } catch (...) {
        logger::writeLog("Configuration file cannot be found", true);
        return "";
    }
}

std::string getGliteLocationFile() {
    std::string filename;
    char * pPath = NULL;
    try {
        pPath = getenv("GLITE_LOCATION");
        if (pPath != NULL) {
            filename = filesStore(DATABASE_FILE_NAME.c_str(), DATABASE_FILE_PATH.c_str(), pPath);
            if (filename.length() > 0)
                return filename;
            else {
                logger::writeLog("GLITE_LOCATION is set, but *.properties file cannot be found under " + DATABASE_FILE_PATH, true);
                return "";
            }

        } else {
            filename = filesStore(DATABASE_FILE_NAME.c_str(), DATABASE_FILE_PATH.c_str(), NULL);
            if (filename.length() > 0)
                return filename;
            else {
                logger::writeLog("GLITE_LOCATION is not set, and not other location contains the *.properties file", true);
                return "";
            }
        }
    } catch (...) {
        logger::writeLog("Database credentials file cannot be found, pls check FTS database *.properties file location", true);
        return "";
    }
}

std::vector<std::string> const& oracleCredentials() {
    std::string word;
    std::string username = "";
    std::string password = "";
    std::string connectstring = "";
    bool userOk = false, passOk = false, connectOk = false;
    int fileFound = -1;
    std::string filename;

    try {
        filename = getGliteLocationFile();
        if (filename.length() == 0)
            return recoveryVector;

        std::ifstream in(filename.c_str());
        if(!in){
	 logger::writeLog("Database credentials file cannot be read, check location and permissions", true);
	 return recoveryVector;
	}
	 
        std::vector<std::string>::iterator it;

        while (in >> word){
            fileDB.push_back(word);
	    }

        for (it = fileDB.begin(); it < fileDB.end(); it++) {
            if (!userOk) {
                username = getUserName(*it, it);
                if (username.length() > 0) {
                    userOk = true;
                    continue;
                }
            }

            if (!passOk) {
                password = getPassword(*it, it);
                if (password.length() > 0) {
                    passOk = true;
                    continue;
                }
            }

            if (!connectOk) {
                connectstring = getConnectString(*it, it);
                if (connectstring.length() > 0) {
                    connectOk = true;
                    continue;
                }
            }
        }

        in.close();
        fileDB.clear();
        fileDB.push_back(username);
        fileDB.push_back(password);
        fileDB.push_back(connectstring);

        return fileDB;
    } catch (...) {
        logger::writeLog("Database credentials file cannot be found", true);
        return recoveryVector;
    }
}

//checks if a string is full of digits

bool isDigits(string word) {
    for (unsigned int i = 0; i < word.size(); ++i) {
        if (!isdigit(word[i])) return false;
    }
    return true;
}

std::string extractNumber(const std::string & value) {
    string sentence = value;
    stringstream extract; // extract words by words;

    extract << sentence; //enter the sentence that we want to extract word by word

    string word = "";

    int totalSum = 0;

    //while there are words to extract
    while (!extract.fail()) {
        extract >> word; //extract the word

        if (isDigits(word)) {
            if (atoi(word.c_str()) > 399 && atoi(word.c_str()) < 554) //report only gridftp error codes			
                return word;
        }
    }
    return "";
}

string strip_space(const string & s) {
    string ret(s);
    while (ret.length() && (ret[0] == ' ' || ret[0] == '\t'))
        ret = ret.substr(1);
    while (ret.length() && (ret[ret.length() - 1] == ' ' || ret[ret.length() - 1] == '\t'))
        ret = ret.substr(0, ret.length() - 1);
    return ret;
}

std::string getMsgConfigFile() {
    std::string filename;
    char * pPath = NULL;
    try {
        pPath = getenv("GLITE_LOCATION");
        if (pPath != NULL) {
            filename = filesStore(CFG_FILE_NAME.c_str(), CFG_FILE_PATH.c_str(), pPath);
            if (filename.length() > 0)
                return filename;
            else {
                logger::writeLog("GLITE_LOCATION is set, but points to incorrect path", true);
                return "";
            }

        } else {
            filename = filesStore(CFG_FILE_NAME.c_str(), CFG_FILE_PATH.c_str(), NULL);
            if (filename.length() > 0)
                return filename;
            else {
                logger::writeLog("GLITE_LOCATION is not set, and no other location contains fts-msg-monitoring.conf file", true);
                return "";
            }

        }
    } catch (...) {
        logger::writeLog("fts-msg-monitoring.conf file cannot be found, pls file location", true);
        return "";
    }
}

std::string getBROKER() {
    return BROKER;
}

std::string getSTART() {
    return START;
}

std::string getCOMPLETE() {
    return COMPLETE;
}

std::string getCRON() {
    return CRON;
}

std::string getTTL() {
    return TTL;
}

bool getTOPIC() {
    return TOPIC;
}

bool getACTIVE() {
    return ACTIVE;
}

bool getENABLELOG() {
    return ENABLELOG;
}

std::string getCRONFQDN()
{
    return CRONFQDN;
}

std::string getLOGFILEDIR()
{
    return LOGFILEDIR;
}

std::string getLOGFILENAME()
{
    return LOGFILENAME;
}

bool getENABLEMSGLOG()
{
    return ENABLEMSGLOG;	
}


std::string getUSERNAME()
{
    return USERNAME;
}

std::string getPASSWORD()
{
   return PASSWORD;
}

bool getUSE_BROKER_CREDENTIALS()
{
   return USE_BROKER_CREDENTIALS;
}


//message broker config file

bool get_mon_cfg_file() {
    std::string boolTOPIC;
    std::string boolACTIVE;
    std::string boolENABLELOG;
    std::string boolENABLEMSGLOG;
    std::string boolUSE_BROKER_CREDENTIALS;    
    std::string filename;

    try {
        filename = getMsgConfigFile();
        if (filename.length() == 0)
            return false;

        string ifname(filename);
        ifstream in(ifname.c_str());
        if(!in){
	 logger::writeLog("msg config file cannot be read, check location and permissions", true);
	 return false;
	}
	 
	cfg.clear();
        string line;
        while (!in.eof()) {
            getline(in, line);	    
            line = strip_space(line);
            if (line.length() && line[0] != '#') {
                int pos = line.find("=");
                if (pos != string::npos) {
                    string key = strip_space(line.substr(0, pos));
                    string value = strip_space(line.substr(pos + 1));
                    cfg.insert(make_pair(key, value));
                }
            }
        }

        map <string, string> ::const_iterator iter1 = cfg.find("BROKER");
        map <string, string> ::const_iterator iter2 = cfg.find("START");
        map <string, string> ::const_iterator iter3 = cfg.find("COMPLETE");
        map <string, string> ::const_iterator iter4 = cfg.find("CRON");
        map <string, string> ::const_iterator iter6 = cfg.find("TTL");
        map <string, string> ::const_iterator iter5 = cfg.find("TOPIC");
        map <string, string> ::const_iterator iter7 = cfg.find("ACTIVE");
        map <string, string> ::const_iterator iter8 = cfg.find("ENABLELOG");
        map <string, string> ::const_iterator iter9 = cfg.find("LOGFILEDIR");
        map <string, string> ::const_iterator iter10 = cfg.find("LOGFILENAME");
        map <string, string> ::const_iterator iter11 = cfg.find("FQDN");
	map <string, string> ::const_iterator iter12 = cfg.find("ENABLEMSGLOG");
	map <string, string> ::const_iterator iter13 = cfg.find("USERNAME");	
	map <string, string> ::const_iterator iter14 = cfg.find("PASSWORD");		
	map <string, string> ::const_iterator iter15 = cfg.find("USE_BROKER_CREDENTIALS");			
	
	
        if (iter13 != cfg.end()) {
            USERNAME = cfg.find("USERNAME")->second;
            if (USERNAME.length() == 0)
		USERNAME = "";
        }	
	
        if (iter14 != cfg.end()) {
            PASSWORD = cfg.find("PASSWORD")->second;
            if (PASSWORD.length() == 0)
		PASSWORD = "";
        }	
		
	
        if (iter15 != cfg.end()) {
            boolUSE_BROKER_CREDENTIALS = cfg.find("USE_BROKER_CREDENTIALS")->second;
            if (boolUSE_BROKER_CREDENTIALS.compare("true") == 0)
                USE_BROKER_CREDENTIALS = true;
            else
                USE_BROKER_CREDENTIALS = false;
        }		
		
        if (iter12 != cfg.end()) {
            boolENABLEMSGLOG = cfg.find("ENABLEMSGLOG")->second;
            if (boolENABLEMSGLOG.compare("true") == 0)
                ENABLEMSGLOG = true;
            else
                ENABLEMSGLOG = false;
        }		
	
        if (iter11 != cfg.end()) {
            CRONFQDN = cfg.find("FQDN")->second;
            if (CRONFQDN.length() == 0)
		CRONFQDN = "";
        }	

        if (iter9 != cfg.end()){
            LOGFILEDIR = cfg.find("LOGFILEDIR")->second;
	    if( LOGFILEDIR[LOGFILEDIR.length()-1] != '/')
	    	LOGFILEDIR += "/";
	    }
        else{
            LOGFILEDIR = "/var/log/glite/";
	    }
	    
	    
        if (iter10 != cfg.end())
            LOGFILENAME = cfg.find("LOGFILENAME")->second;
        else
            LOGFILENAME = "msg.log";


        if (iter1 != cfg.end())
            BROKER = cfg.find("BROKER")->second;
        if (iter2 != cfg.end())
            START = cfg.find("START")->second;
        if (iter3 != cfg.end())
            COMPLETE = cfg.find("COMPLETE")->second;
        if (iter4 != cfg.end())
            CRON = cfg.find("CRON")->second;
        if (iter6 != cfg.end())
            TTL = cfg.find("TTL")->second;

        if (iter5 != cfg.end()) {
            boolTOPIC = cfg.find("TOPIC")->second;
            if (boolTOPIC.compare("true") == 0)
                TOPIC = true;
            else
                TOPIC = false;
        }
        if (iter7 != cfg.end()) {
            boolACTIVE = cfg.find("ACTIVE")->second;
            if (boolACTIVE.compare("true") == 0)
                ACTIVE = true;
            else
                ACTIVE = false;
        }
        if (iter8 != cfg.end()) {
            boolENABLELOG = cfg.find("ENABLELOG")->second;
            if (boolENABLELOG.compare("true") == 0)
                ENABLELOG = true;
            else
                ENABLELOG = false;
        }

        return true;
    } catch (...) {
        logger::writeLog("msg config file error", true);
        return false;
    }
}

void appendMessageToLogFile(std::string & text) {
    static std::string filename = LOGFILEDIR + "" + LOGFILENAME;
    static ofstream fout;   

        fout.open(filename.c_str(), ios::app); // open file for appending
        if (fout.is_open()){
            fout << text << endl; //send to file
	}
	fout.flush();
        fout.close(); //close file
}

int GetIntVal(std::string strConvert) {
    try {
        if (isDigits(strConvert))
            return (atoi(strConvert.c_str())) * MILLI;

        return MILLI; //default value is ten hours
    } catch (...) {
        return MILLI; //default value is ten hours
    }

}

std::string ReplaceNonPrintableCharacters(string s) {
    try {
        std::string result;
        for (int i = 0; i < s.length(); i++) {
            char c = s[i];
            int AsciiValue = static_cast<int> (c);
            if (AsciiValue < 32 || AsciiValue > 127) {
                result.append(" ");
            } else {
                result += s.at(i);
            }
        }
        return result;
    } catch (...) {
        return s;
    }
}


bool caseInsCompare(const string& s1, const string& s2) {
   return((s1.size( ) == s2.size( )) &&
          equal(s1.begin( ), s1.end( ), s2.begin( ), caseInsCharCompareN));
}


bool send_message(std::string & text)
{   
       bool msgSent = true;       
       int fd = 0;
       int fd2 = 0;
       int fd3 = 0;
       
       size_t writen;

       //ignore SIGPIPE signal, will be handled by EPIPE errno
       sigignore(SIGPIPE);
       
       //open first named pipe
       if( (fd = open(HALF_DUPLEX, O_WRONLY  | O_NONBLOCK)) == -1 ) { 
          msgSent = false;	  
       }
       
            //open second named pipe
       if( (fd2 = open(HALF_DUPLEX2, O_WRONLY  | O_NONBLOCK)) == -1 ) {
          msgSent = false;
       }

            //open third named pipe
       if( (fd3 = open(HALF_DUPLEX3, O_WRONLY  | O_NONBLOCK)) == -1 ) {
          msgSent = false;
       }

     	writen = write(fd, text.c_str(), text.length());
	
   if(  ((int)writen < text.length()) ||   (writen == -1) || (errno==EPIPE) || (errno==EAGAIN)  )  { 
	   errno = 0; //reset it 
           writen =  write(fd2, text.c_str(), text.length());              
           if(  ((int)writen < text.length()) ||   (writen == -1) || (errno==EPIPE) || (errno==EAGAIN)  )  {
             errno = 0; //rest it again
             writen =  write(fd3, text.c_str(), text.length());
               if(  ((int)writen < text.length()) ||   (writen == -1) || (errno==EPIPE) || (errno==EAGAIN)  )  {
                 msgSent = false;
            }
          }
	}

      if( ((int)writen == text.length())){        
         msgSent = true;
	 }
 
       if(fd != -1)
        close(fd);
       if(fd2 != -1)
        close(fd2);     
       if(fd3 != -1)
        close(fd3);

  return msgSent;	
}


bool getResolveAlias()
{
  struct hostent *he;
  struct in_addr a;
  std::string fqdn =  getCRONFQDN(); //from config file
  char hostname[512] = {0};
  gethostname(hostname, 512);
  std::string currenthost = hostname; //current hostname
  std::string tempAlias;
  
  he = gethostbyname (hostname);
  if (he){
    std::string tempHostname = he->h_name;
    if(true == caseInsCompare(tempHostname, fqdn)){
    	return true;
    }    
     while (*he->h_aliases){
	    tempAlias = *he->h_aliases++;
	    if(true == caseInsCompare(tempAlias, fqdn)){
    		return true;
	    }
    }    

     }
  return false;
}  

std::string getFTSEndpoint()
{
const char* path[2] = {"/etc/glite-sd2cache-cron.conf","/opt/glite/etc/glite-sd2cache-cron.conf"};
int rval = 0;
string line;
vector<std::string>::iterator myVectorIterator;
size_t found;
std::string fts = "";
std::string result = "";
 
 for(int index=0; index<2; index++){
        string ifname(path[index]);
        ifstream in(ifname.c_str());
        if(!in){
	 continue;
	}
	 
        string line;
        while (!in.eof()) {
            getline(in, line);	    
            line = strip_space(line);
            if (line.length() && line[0] != '#') {
                int pos = line.find("=");
                if (pos != string::npos) {
                    string key = strip_space(line.substr(0, pos));
                    string value = strip_space(line.substr(pos + 1));
                    ftsendpoint.insert(make_pair(key, value));
                }
            }
        }

        map <string, string> ::const_iterator iter = ftsendpoint.find("FTS_HOST"); 
	if (iter != ftsendpoint.end()) {
            fts = ftsendpoint.find("FTS_HOST")->second;
            if (fts.length() == 0)
		fts = "";
        }	
  }
  if(fts.length() > 0)
  {
    fts.erase(0, 1);
    fts.erase(fts.length()-1, fts.length());
    result = "https://";
    result += fts;
    result += ":8443/glite-data-transfer-fts/services/FileTransfer";
  }
  return result;
}
